#include "draw_buffers.h"
#include "draw_primitives.h"
#include "graph_vram.h"
#include "gs_psm.h"
#include <tamtypes.h>
#include <sifrpc.h>
#include <debug.h>
#include <unistd.h>
#include <dma.h>
#include <graph.h>
#include <draw.h>
#include <stdio.h>
#include <inttypes.h>
#include <gs_gp.h>
#include <stdlib.h>

#define myftoi4(x) ((x)<<4)

#ifdef DEBUG
#define DEBUG_MSG(...) printf(__VA_ARGS__)
#else
#define DEBUG_MSG(...)
#endif


#define OFFSET_X 2048
#define OFFSET_Y 2048

#define VID_W 640
#define VID_H 448

int print_buffer(qword_t *b, int len) {
    DEBUG_MSG("-- buffer\n");
    for(int i = 0; i < len; i++) {
        DEBUG_MSG("%016llx %016llx\n", b->dw[0], b->dw[1]);
        b++;
    }
    DEBUG_MSG("-- /buffer\n");
    return 0;
}

int gs_finish() {
    qword_t buff[50];
    qword_t *q = buff;
    //TODO: Remove!
    q = draw_finish(q);
    dma_channel_send_normal(DMA_CHANNEL_GIF, buff, q-buff, 0, 0);
    dma_wait_fast();
    return 0;
}
zbuffer_t *z;
int gs_init(int width, int height, int psm, int psmz, int vmode, int gmode) {
    framebuffer_t fb;
    fb.address = graph_vram_allocate(width, height, psm, GRAPH_ALIGN_PAGE);
    fb.width = width;
    fb.height = height;
    fb.psm = psm;
    fb.mask = 0;
    z->address = graph_vram_allocate(width, height, psmz, GRAPH_ALIGN_PAGE);
    z->enable = 0;
    z->method = 0;
    z->zsm = 0;
    z->mask = 0;
    graph_set_mode(gmode, vmode, GRAPH_MODE_FIELD, GRAPH_DISABLE);
    graph_set_screen(0, 0, width, height);
    graph_set_bgcolor(0, 0, 0);
    graph_set_framebuffer_filtered(fb.address, width, psm, 0, 0);
    graph_enable_output();
    qword_t buf[100];
    qword_t *q = buf;
    q = draw_setup_environment(q, 0, &fb, z);
    q = draw_primitive_xyoffset(q, 0, 2048-(VID_W/2), 2048-(VID_H/2));
    q = draw_finish(q);
    dma_channel_send_normal(DMA_CHANNEL_GIF, buf, q-buf, 0, 0);
    dma_wait_fast();
    return 0;
}

static int verts[] = {
    100, 100, 0,
    300, 100, 0,
    100, 300, 0,
    300, 100, 0,
    300, 300, 0,
    100, 300, 0
};

#define SHIFT_AS_I64(x, b) (((uint64_t)x)<<b);

uint64_t red = 0xf0;
uint64_t blue = 0x0f;
uint64_t green = 0xf0;
int draw() {


    qword_t buf[50];
    qword_t *q = buf;
    q = draw_disable_tests(q, 0, z);
    q = draw_clear(q, 0, 2048.0f - 320, 2048.0f - 244, VID_W, VID_H, 20, 20, 20);
    //q = draw_enable_tests(q, 0, z);
    // GIFTag header - 3x(col, pos)
    q->dw[0] = 0xD000000000000001;
    q->dw[1] = 0x0005151515151510;
    q++;

    q->dw[0] = GS_PRIM_TRIANGLE;
    q->dw[1] = 0;
    q++;

    int cx = myftoi4(2048 - (VID_W/2));
    int cy = myftoi4(2048 - (VID_H/2));

    for(int i = 0; i < 6; i++){
        q->dw[0] = (red&0xff) | (green&0xff)<<32;
        q->dw[1] = (blue&0xff) | SHIFT_AS_I64(0x80, 32);
        q++;

        int base = i*3;
        int x = myftoi4(verts[base+0]) + cx;
        int y = myftoi4(verts[base+1]) + cy;
        int z = myftoi4(verts[base+2]);
        q->dw[0] = x | SHIFT_AS_I64(y, 32);
        q->dw[1] = z; 
        q++;
    }
    q = draw_finish(q);
    print_buffer(buf, q-buf);
    dma_channel_send_normal(DMA_CHANNEL_GIF, buf, q-buf, 0, 0);
    dma_wait_fast();
    //gs_finish();
    return 0;
}

int main() {
#ifdef DEBUG
    sceSifInitRpc(0);
    init_scr();
#endif
    DEBUG_MSG("PS2 Initialized\n");
    // initialize DMAC
    dma_channel_initialize(DMA_CHANNEL_GIF, 0, 0);
    dma_channel_fast_waits(DMA_CHANNEL_GIF);
    DEBUG_MSG("DMA Initialized...\n");
    // initialize graphics mode
    gs_init(VID_W, VID_H, GS_PSM_32, GS_PSMZ_24, GRAPH_MODE_NTSC, GRAPH_MODE_INTERLACED);
    DEBUG_MSG("GS Initialized...\n");
    // clear screen

    // build buffer with triangle data

    z = malloc(sizeof(zbuffer_t));

    while(1) {
        red++;
        if(red > 0xFF) {
            red = 0;
        }
        draw();
        graph_wait_vsync();
    }
}